/* "fgrind.c"	K. J. Turner <kjt@cs.stir.ac.uk>	23/09/97
 *
 * The main file for "fgrind", a FrameMaker prettyprinter
 *
 * Produced by Ken Turner, University of Stirling (kjt@@cs.stir.ac.uk).
 * Heavily based on the original for "lgrind" (version 1.4) by Van Jacobson
 * Apart from changes to produce MML, the following modifications were
 * carried out:
 *
 * 27/10/94:
 *
 *   o removed options "-h" (header), "-i" (included document)
 *
 *   o changed option "-d" (definitions) to give directory only
 *
 *   o added options "-g" (line number gap), "-s" (point size)
 *
 * 23/09/97:
 *
 *   o "char" changed to "unsigned char" in most cases to allow for 8-bit
 *      characters
 *
 *   o translation table changed to deal with NEXTSTEP 8-bit characters
 *     (change "printtab" to deal with your local character set if you want
 *     to use characters 128 to 255)
 *
 * Much of the original code dealt with (La)TeX text embedded in special
 * delimiters. This was half-heartedly converted to produce MML paragraphs
 * but is unlikely to make much sense. The main purpose of "fgrind" is to
 * turn a program file into an MML file.
 *
 * Original of "lgrind" copyright 1985 by Van Jacobson (van@@lbl-rtsg.arpa),
 * Lawrence Berkeley Laboratory, who included the following remarks.
 *
 * This program may be freely used and copied but may not be sold
 * without the author's written permission.  This notice must remain
 * in any copy or derivative.
 *
 * This program is an adaptation of "vfontedpr" v4.2 (12/11/84) from
 * the 4.2bsd Unix distribution.  Vfontedpr was written by Dave
 * Presotto (based on an earlier program of the same name written by
 * Bill Joy).
 *
 */
   
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define DEFS "fgrindefs"
#define LMARGIN 0.3

#ifdef NeXT
#define MMLEXT ".framemml"
#else
#define MMLEXT ".mml"
#endif

#ifndef vms
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <time.h>
#else
#  include <types.h>
#  include <stat.h>
#endif
   
typedef	int boolean;

#define TRUE  1
#define FALSE 0

#define STANDARD 0
#define ALTERNATE 1
   
#define	NOTCODE 0		/* the three states of incode */
#define INITCODE 1
#define OTHERCODE 2

#define PNAMELEN 80		/* length of a function/procedure name */
#define PSMAX 40		/* size of procedure name stacking */
   
#define	BUFFERSIZE	4096

#ifndef vms
#  define OKEXIT	0
#  define BADEXIT	1
#  ifndef DEFSDIR
#    define DEFSDIR	"/usr/local/lib"
#  endif
#else
#  define OKEXIT	1
#  define BADEXIT	0
#  define DEFSDIR	"TEX$INPUTS"
#endif
   

/* regular expression routines */
   
char	*expmatch();		/* match a string to an expression */
char	*convexp();		/* convert expression to internal form */


/* database query routines */

char	*tgetstr();
int	 tgetent();
int	 tgetflag();


/* forward declarations */

void	setlang();
void	readfile();
void	putScp();
void	putKcp();
void	putVcp();
void	putstr();
boolean	isproc();
void	outchar();
void	mk_header();
char	*out_fname();

/*
 *	The state variables
 */

boolean	incomm;			/* in a comment of the primary type */
boolean	instr;			/* in a string constant */
boolean	inchr;			/* in a character constant */
int	incode;			/* in program text within a comment */
boolean	use_tt;			/* use \tt everywhere */
boolean	do_at;			/* pay attention to ats in Frame */
boolean	do_tt;			/* pay attention to bars in Frame */
boolean	nokeyw = FALSE;		/* no keywords being flagged */
boolean	prccont;		/* continue last procedure */
boolean	code_cmnts = TRUE;	/* Treat at, etc specially in comments */
boolean	linehead;		/* at beginning of line or after whitespace */
int	fsize = 10;		/* font size */
int	linegap = 0;		/* line number gap (0 = no numbers) */
int	gap;			/* size of gap in text */
int	tabcol;			/* last column tabbed to (0 = start) */
int	spacecol;		/* next column to be tabbed to (0 = start) */
int	lastout;		/* (extended) last character to outchar */
int	comtype;		/* type of comment */
int	psptr;			/* the stack index of the current procedure */
char	pstack[PSMAX][PNAMELEN+1]; /* the procedure name stack */
int	plstack[PSMAX];		/* the procedure nesting level stack */
int	blklevel;		/* current nesting level */
char	*fname = "";		/* File being read */
int	lineno;			/* Line number in current file */
char	pname[BUFFERSIZE+1];	/* Current procedure name */
char	*progname;		/* Program name */

/*
 *	The language specific globals
 */

unsigned char *defsdir = DEFSDIR; /* name of language definitions directory */
unsigned char *defsfile;	/* name of language definitions file */
unsigned char language[PNAMELEN]="c"; /* the language indicator */
unsigned char the_buf[BUFFERSIZE+1]; /* general purpose buffer */
unsigned char *buf = the_buf + 1; /* buf-1 must be valid */
unsigned char strings[2*BUFFERSIZE]; /* store the keywords */
unsigned char defs[2 * BUFFERSIZE]; /* language definitions */
unsigned char *l_keywds[BUFFERSIZE/2]; /* keyword table address */
unsigned char *l_prcbeg;	/* regular expr for procedure begin */
unsigned char *l_combeg;	/* regexp introducing a comment */
unsigned char *l_comend;	/* regexp ending a comment */
unsigned char *l_acmbeg;	/* regexp introducing a comment */
unsigned char *l_acmend;	/* regexp ending a comment */
unsigned char *l_blkbeg;	/* regexp beginning of a block */
unsigned char *l_blkend;	/* regexp ending a block */
unsigned char *l_strbeg;	/* regexp starting string constant */
unsigned char *l_strend;	/* regexp ending string constant */
unsigned char *l_chrbeg;	/* regexp starting char. constant */
unsigned char *l_chrend;	/* regexp ending character constant */
unsigned char *l_cdebeg;	/* regexp starting prog text in comment */
unsigned char s_cdebeg[PNAMELEN+1]; /* actual string corr. to l_cdebeg */
unsigned char *l_cdeend;	/* regexp ending prog text in comment */
unsigned char *l_texbeg;	/* regexp starting Frame text in comment */
unsigned char *l_texend;	/* regexp ending Frame text in comment */
unsigned char *l_txmbeg;	/* regexp starting Frame math in comment */
unsigned char *l_txmend;	/* regexp ending Frame math in comment */
unsigned char *l_tt_beg;	/* regexp starting verbatim text in comment */
unsigned char *l_tt_end;	/* regexp ending typewriter text in comment */
unsigned char *l_at;		/* regexp for at in Frame text */
unsigned char *l_tt;		/* regexp for bar in Frame text */
unsigned char *l_pc;		/* regexp for percent in Frame text */
unsigned char l_escape;		/* character used to escape characters */
boolean	l_toplex;		/* procedures only defined at top lex level */
boolean	l_onecase;		/* upper & lower case equivalent */
boolean embed = FALSE;          /* -e seen --- do embedded code */
boolean	code = TRUE;		/* Looking at code */

/*
 *  global variables also used by expmatch
 */

extern	boolean	_escaped;	/* if last character was an escape */
extern	unsigned char *_start;	/* start of the current string */

extern	int (*re_strncmp)();	/* function to do string compares */
extern	int strncmp();
extern	int lc_strncmp();

   int
main(argc, argv)
   int	argc;
   char	*argv[];
{
   struct stat	stbuf;
   char		*p;
   char		*ofname;
   int		files_done = 0;
   int		i;
   
   argc = getredirection(argc, argv);
   progname = (char *)malloc(strlen(argv[0]+1));
   strcpy(progname, argv[0]);
   if (argc == 1) {
     fprintf(stderr, "usage: %s [-c] [-ddefdir] [-ggap] [-llang]", progname);
     fprintf(stderr, "[-n] [-ssize] [--|file...]\n");
     exit(BADEXIT);
   }
   argc--, argv++;
   
   do {
      register char *cp;
      register int i;
      
      if (argc > 0) {
	 /* read from standard input */
	 if (!strcmp(argv[0], "-")) {
	    argc = 0;
	    goto rest;
	 }
	 
	 /* Do not treat at, etc. specially in comments */
	 if (!strcmp(argv[0], "-c")) {
	    code_cmnts = FALSE;
	    argc--, argv++;
	    continue;
	 }
	 
	 /* Do treat at, etc. specially in comments */
	 if (!strcmp(argv[0], "+c")) {
	    code_cmnts = TRUE;
	    argc--, argv++;
	    continue;
	 }
	 
	 /* specify the language description directory */
	 if (!strncmp(argv[0], "-d", 2)) {
	    defsdir = (char *) malloc(strlen(argv[0]+2)+1);
	    strcpy(defsdir, argv[0]+2);
	    argc--, argv++;
	    continue;
	 }
	 
	 /* process embedded text */
	 if (!strcmp(argv[0], "-e")) {
	    embed = TRUE;
	    argc--, argv++;
	    continue;
	 }

	 /* define line number gap */
	 if (!strncmp(argv[0], "-g", 2)) {
	    i = atoi(argv[0]+2);
	    if (i > 0)
	      linegap = i;
	    else {
	      fprintf(stderr, "%s: line number gap invalid\n", progname);
	    }
	    argc--, argv++;
	    continue;
	 }
	 
	 /* specify the language */
	 if (!strncmp(argv[0], "-l", 2)) {
	    strcpy(language, argv[0]+2);
	    argc--, argv++;
	    continue;
	 }
	 
	 /* indicate no keywords */
	 if (!strcmp(argv[0], "-n")) {
	    nokeyw++;
	    argc--, argv++;
	    continue;
	 }
	 
	 /* define point size */
	 if (!strncmp(argv[0], "-s", 2)) {
	    i = atoi(argv[0]+2);
	    if (i > 0)
	      fsize = i;
	    else {
	      fprintf(stderr, "%s: point size invalid\n", progname);
	    }
	    argc--, argv++;
	    continue;
	 }
	 
	 /* Use \tt for all fonts */
	 if (!strcmp(argv[0], "-t")) {
	    use_tt = TRUE;
	    argc--, argv++;
	    continue;
	 }
	 
	 /* open the file for standard input */
	 if (freopen(argv[0], "r", stdin) == NULL) {
	    perror(argv[0]);
	    exit(BADEXIT);
	 }
	 
	 fname = argv[0];
	 lineno = 0;
	 argc--, argv++;
      }

   rest:
      /* initialize the program */
      
      if (files_done == 0) {
	/* open the file for standard output */
	if (strcmp(fname, "") != 0) {
	  ofname = out_fname(fname);
	  if (freopen(ofname, "w", stdout) == NULL) {
	    perror(ofname);
	    exit(BADEXIT);
	  }
	}
	mk_header(fsize);
      }
	 
      defsfile = (char *) malloc(strlen(defsdir)+11);
      strcpy(defsfile, defsdir);
      strcat(defsfile, "/");
      strcat(defsfile, DEFS);

      setlang();

      if (strcmp(fname, "") != 0) {
        printf("<Comment \"%s\" (version of ", fname);
        fstat(fileno(stdin), &stbuf);
        cp = ctime(&stbuf.st_mtime);
        cp[7] = cp[10] = cp[16] = cp[24] = '\0';
        printf("%s, %s %s %s)>\n\n", cp+11, cp+8, cp+4, cp+20);
      }

      incomm = instr = inchr = _escaped = FALSE;
      incode = NOTCODE;
      do_at = do_tt = TRUE;
      blklevel = 0;
      for (psptr = 0; psptr < PSMAX; psptr++) {
	 pstack[psptr][0] = '\0';
	 plstack[psptr] = 0;
      }
      psptr = -1;
      
      code = FALSE;
      readfile(stdin);
      printf("\n");
      files_done++;

      if (code)
	 fprintf(stderr, "%s: reached EOF within code in file %s\n", progname,
	   fname);
   } while (argc > 0);
   
   exit(OKEXIT);
}

/*
 * out_fname --- return output filename based on input filename "fname"
 */
   char *
out_fname(fname)
   char *fname;
{
   static char ofname[80];
   char *dotptr, *slashptr;
   
   strcpy(ofname, fname);
   dotptr = strrchr(ofname, '.');
   slashptr = strrchr(ofname, '/');
   if (dotptr != NULL && dotptr > slashptr)
     *dotptr = '\0';
   strcat(ofname, MMLEXT);
   return(ofname);
}

/*
 * mk_header --- output MML file header for font size "fsize"
 */
   void
mk_header(fsize)
   int fsize;
{
  int i;
  float tsize;
  
  printf("<MML - fgrind listing>\n\n");
  printf("<!DefineMacro TS \"<Pts %d>\">\n", fsize);
  printf("<!DefineMacro LS \"<Pts %d>\">\n", fsize - 2);
  tsize = fsize * 0.012;
  for (i = 0; i < 40; i++)
    printf("<!DefineMacro T%02d \"<TabStop %1.2f>\">\n", i,
      (i + 1) * tsize + LMARGIN);
  printf("\n");
  printf("<Include %s/%s.mml>\n\n", defsdir, progname);
  printf("<Program><N>\n\n");
}

/*
 * readfile() --- read and process a file
 */
   void
readfile(fp)
   FILE *fp;
{
   register unsigned char *cp;
   boolean Inline = FALSE;		/* Doing %( -- %) */
   unsigned char term;			/* ] or ) */
   unsigned char *atptr;		/* start of at within Frame text */
   unsigned char *atendptr;		/* end of at within Frame text */
   unsigned char *ttptr;		/* start of bar within Frame text */
   unsigned char *ttendptr;		/* end of bar within Frame text */
   unsigned char *pcptr;		/* start of percent in Frame text */
   unsigned char *pcendptr;		/* end of percent within Frame text */
   unsigned char temp[BUFFERSIZE];
   
   while (fgets(buf, BUFFERSIZE, fp) != NULL) {
      lineno++;
      cp = buf;
      linehead = TRUE;
      lastout = '\0';
      if (embed) {
	 if (!code) {
	    if (buf[0] == '%' &&
		(buf[1] == '[' || buf[1] == '('
		 || buf[1] == '#' || buf[1] == '<' || buf[1] == '!'
		 || buf[1] == '@' || buf[1] == '|')) {
	       
	       for (cp = buf + 2; *cp == ' ' || *cp == '\t'; cp++)
		  ;

	       /* Change the language */
	       if (buf[1] == '#') {
		  if (*cp == '\n')
		     fprintf(stderr, "%s: no language seen after %%#\n",
		       progname);
		  else {
		     cp[strlen(cp) - 1] = '\0';	/* nuke the '\n' */
		     strcpy(language, cp);
		     printf("%% switching to %s\n", language);
		     setlang();
		  }
		  continue;
	       }
	       
	       /* Turn at or bar processing within Frame
	        * text on or off. */

	       if (buf[1] == '@' || buf[1] == '|') {
		  if (*cp == '\n')
		     fprintf(stderr, "%s: no setting seen after %%%c\n",
		       progname, buf[1]);
		  else {
		     int flag = (*cp == '1' || *cp == '+');
		     if (buf[1] == '@') {
			do_at = flag;
			}
		     else
			do_tt = flag;
		  }
		  continue;
	       }
	       
	       code = TRUE;
	       
	       /* take input from another file or from a shell command */
	       if (buf[1] == '<' || buf[1] == '!') {
		  FILE *fp;
		  unsigned char source = buf[1];

		  cp[strlen(cp) - 1] = '\0';	/* nuke the '\n' */
		  if (source == '<')
		     fp = fopen(cp, "r");
		  else /* source == '!' */
		     fp = popen(cp, "r");

		  if (fp == NULL) {
		     sprintf(temp, "%%%c on `%s' failed", source, cp);
		     perror(temp);
		  } else {
		     strcpy(temp, cp);
		     printf("<Comment Start of `%s'>\n", temp);
		     printf("<Program>\n");
		     readfile(fp);
		     printf("\n");
		     printf("<Comment End of `%s'>\n", temp);
		     if (source == '<')
			fclose(fp);
		     else
			pclose(fp);
		  }
		  code = FALSE;
		  continue;
	       }
	       
	       /* embedded inline or displayed code */
	       if (buf[1] == '(') {
		  Inline = TRUE;
		  term = ')';
	       } else {	/* buf[1] == '[' */
		  Inline = FALSE;
		  term = ']';
	       }
	       if (buf[2] == '\n') {
		  printf("<Program>\n");
	       } else {
		  buf[strlen(buf)-1] = '\0';
		  printf("%s<Program>\n", &buf[2]);
	       }
	       continue;

	    } else if (do_at &&
		       (atendptr = expmatch(buf, l_at, &atptr,
					    (char *) NULL)) != NULL) {
	       /* This is a gross hack.  Fix it, dammit */
	    restart_at:
	       pcendptr = expmatch(cp, l_pc, &pcptr, (char *) NULL);

	       if (pcptr == NULL || atptr < pcptr) {
		  putVcp(cp, atptr-1);
		  printf("<Program>\n");
		  cp = atendptr;
		  atendptr = expmatch(cp, l_at, &atptr, (char *) NULL);
		  if (atptr != NULL && (pcptr == NULL || atptr < pcptr)) {
		     strncpy(temp, cp, atptr-cp);
		     temp[atptr-cp] = '\0';
		     cp = atendptr;
		     putScp(temp);
		     printf("\n");
		     atendptr = expmatch(cp, l_at, &atptr, (char *) NULL);
		     if (atptr != NULL)
			goto restart_at;
		     else {
			fputs(cp, stdout); continue;
		     }
		  } else {
		     /* multiline */
		  }
	       } else {
		  fputs(buf, stdout);
		  continue;
	       }
	    } else if (do_tt &&
		       (ttendptr = expmatch(buf, l_tt, &ttptr,
					    (char *) NULL)) != NULL) {
	       pcendptr = expmatch(buf, l_pc, &pcptr, (char *) NULL);

	       if (pcptr == NULL || ttptr < pcptr) {
		  putVcp(buf, ttptr-1);
	       } else {
		  fputs(buf, stdout); continue;
	       }
	    } else {
	       fputs(buf, stdout);
	       continue;
	    }
	 } /* if (!code) */
	 
	 /*
	  * We're in embedded code.
	  */
	 if (buf[0] == '%') {
	    if (buf[1] == '*') {
	       if (term != ']')
		  fprintf(stderr,
		    "%s: %%* only makes sense in display mode\n", progname);
	       else {
		  printf("\n");
		  printf("<Program>\n");
	       }
	       continue;
	    } else if (buf[1] == term) {
	       if (buf[2] == '\n')
		  printf("\n");
	       else
		  printf("\n%s", &buf[2]);
	       code = FALSE;
	       continue;
	    } else if (buf[1] == '=') {	/* Send literal	*/
	       fputs(&buf[2], stdout);
	       continue;
	    } else if (buf[1] == '[' || buf[1] == '('
		       || buf[1] == ']' || buf[1] == ')') {
	       fprintf(stderr, "%s: possible nested embedded code at line %d in %s\n",
	         progname, lineno, fname);
	    }
	 }
	 
	 /*
	  * Inline code --- suppress leading whitespace
	  */
	 if (Inline) {
	    while (isspace(*cp))
	       cp++;
	 }
      } /* if (embed) */
      
      if (*cp == '\f') {
	 printf("\n<NewPage>\n");
	 cp++;
	 if (*cp == '\n')  /* some people like ^Ls on their own line */
	    continue;
      }
      prccont = FALSE;
      putScp(cp);
      if (!embed && prccont && (psptr >= 0)) {
	 printf("<Comment ");
	 putstr(pstack[psptr]);
	 printf(" Continued>");
      }
#ifdef DEBUG
      printf("com %o str %o chr %o ptr %d\n", incomm, instr, inchr, psptr);
#endif
   } /* while fgets() */
}



/*
 * Set all of the language-dependent variables
 */
   void
setlang()
{
   char	*p, *cp;
   int i;

   /*
    *  get the language definition from the defs file
    */
   i = tgetent(defs, language, defsfile);
   if (i == 0) {
      fprintf(stderr, "%s: no entry for language %s\n", progname, language);
      exit(BADEXIT);
   } else if (i < 0) {
      fprintf(stderr,  "%s: cannot find definitions file `%s'\n", progname,
        defsfile);
      exit(BADEXIT);
   }

   p = strings;
   if (tgetstr("kw", &p) == NULL)
      nokeyw = TRUE;
   else  {
      unsigned char **cpp;
      
      cpp = l_keywds;
      cp = strings;
      while (*cp) {
	 while (*cp == ' ' || *cp =='\t')
	    *cp++ = '\0';
	 if (*cp)
	    *cpp++ = cp;
	 while (*cp != ' ' && *cp  != '\t' && *cp)
	    cp++;
      }
      *cpp = NULL;
   }
   
   p = buf;  l_prcbeg = convexp(tgetstr("pb", &p));
   p = buf;  l_combeg = convexp(tgetstr("cb", &p));
   p = buf;  l_comend = convexp(tgetstr("ce", &p));
   p = buf;  l_acmbeg = convexp(tgetstr("ab", &p));
   p = buf;  l_acmend = convexp(tgetstr("ae", &p));
   p = buf;  l_strbeg = convexp(tgetstr("sb", &p));
   p = buf;  l_strend = convexp(tgetstr("se", &p));
   p = buf;  l_blkbeg = convexp(tgetstr("bb", &p));
   p = buf;  l_blkend = convexp(tgetstr("be", &p));
   p = buf;  l_chrbeg = convexp(tgetstr("lb", &p));
   p = buf;  l_chrend = convexp(tgetstr("le", &p));
   p = s_cdebeg;  l_cdebeg = convexp(tgetstr("zb", &p));
   p = buf;  l_cdeend = convexp(tgetstr("ze", &p));
   p = buf;  l_texbeg = convexp(tgetstr("tb", &p));
   p = buf;  l_texend = convexp(tgetstr("te", &p));
   p = buf;  l_txmbeg = convexp(tgetstr("mb", &p));
   p = buf;  l_txmend = convexp(tgetstr("me", &p));
   p = buf;  l_tt_beg = convexp(tgetstr("vb", &p));
   p = buf;  l_tt_end = convexp(tgetstr("ve", &p));

   l_tt = convexp("\\|");
   l_at = convexp("@");
   l_pc = convexp("%");
   
   l_escape = '\\';
   l_onecase = tgetflag("oc");
   
   re_strncmp = l_onecase ? lc_strncmp : strncmp;
   l_toplex = tgetflag("tl");
}

/*
 * Write out a formatted line of program text
 */
   void
putScp(os)
   unsigned char	*os;
{
   register unsigned char *s = os; /* pointer to unmatched string */
   register unsigned char *s1;	/* temp. string */
   unsigned char *comptr;	/* start of a comment delimiter */
   unsigned char *comendptr;	/* end of a comment delimiter */
   unsigned char *acmptr;	/* start of an alt. comment delimiter */
   unsigned char *acmendptr;	/* end of an alt. comment delimiter */
   unsigned char *strptr;	/* start of a string delimiter */
   unsigned char *strendptr;	/* end of a string delimiter */
   unsigned char *chrptr;	/* start of a char. const delimiter */
   unsigned char *chrendptr;	/* end of a char. const delimiter */
   unsigned char *cdeptr;	/* start of prog text delim in a comment */
   unsigned char *cdeendptr;	/* end of prog text delim in a comment */
   unsigned char *cdbptr;	/* start of prog text delim in a comment */
   unsigned char *cdbendptr;	/* end of prog text delim in a comment */
   unsigned char *texptr;	/* start of Frame text delim in a comment */
   unsigned char *texendptr;	/* end of Frame text delim in a comment */
   unsigned char *txmptr;	/* start of Frame math delim in a comment */
   unsigned char *txmendptr;	/* end of Frame math delim in a comment */
   unsigned char *tt_ptr;	/* start of typewriter delim in a comment */
   unsigned char *tt_endptr;	/* end of typewriter delim in a comment */
   unsigned char *blksptr;	/* start of a lexical block start */
   unsigned char *blksendptr;	/* end of a lexical block start */
   unsigned char *blkeptr;	/* start of a lexical block end */
   unsigned char *blkeendptr;	/* end of a lexical block end */
   
   gap = 0;		/* no current gap */
   tabcol = 0;		/* no tabs issued yet */
   spacecol = 0;	/* no column to be tabbed to yet */
   _start = os;		/* remember the start for expmatch() */
   _escaped = FALSE;
   
   if (linegap && (lineno % linegap) == 0)
     printf("<T><L>%d<N><T>", lineno);
   else
     printf("<T><T>");
   if (nokeyw || incomm || instr)
      goto skip;
   if (isproc(s)) {
      printf("<Comment ");
      putstr(pname);
      printf(">");
      if (!embed && psptr < PSMAX-1) {
	 ++psptr;
	 strncpy(pstack[psptr], pname, PNAMELEN);
	 pstack[psptr][PNAMELEN] = '\0';
	 plstack[psptr] = blklevel;
      }
   } 
 skip:
   do {
      /* check for string, comment, blockstart, etc */
      if (!incomm && !instr && !inchr) {
	 
	 blkeendptr = expmatch(s, l_blkend, &blkeptr, (char *) NULL);
	 blksendptr = expmatch(s, l_blkbeg, &blksptr, (char *) NULL);
	 comendptr =  expmatch(s, l_combeg, &comptr,  (char *) NULL);
	 acmendptr =  expmatch(s, l_acmbeg, &acmptr,  (char *) NULL);
	 strendptr =  expmatch(s, l_strbeg, &strptr,  (char *) NULL);
	 chrendptr =  expmatch(s, l_chrbeg, &chrptr,  (char *) NULL);
	 
	 /* check for end of program text in comment */
	 if (incode != NOTCODE) {
	    cdeendptr = expmatch(s, l_cdeend, &cdeptr, (char *) NULL);

	    if ((cdeptr == NULL && (comptr != NULL || acmptr != NULL))
		|| (cdeptr != NULL
		    && ((comptr != NULL && comptr < cdeptr)
			|| ((acmptr != NULL && acmptr < cdeptr))))) {
	       fprintf(stderr,
                 "%s: comments may not be nested within program text within comments\n",
		   progname);
	       s = (comptr != NULL) ? comptr : acmptr;
	       incode = NOTCODE;
	       continue;
	    }

	    /* look to see if there's a \<zb> in the program text */
	    cdbendptr = expmatch(s, l_cdebeg, &cdbptr, (char *) NULL);
	    if (cdeptr != NULL
		&& (strptr  == NULL || cdbptr < strptr)
		&& (chrptr  == NULL || cdbptr < chrptr)
		&& (blksptr == NULL || cdbptr < blksptr)
		&& (blkeptr == NULL || cdbptr < blkeptr)) {
	       if (cdbptr > s && cdbptr[-1] == l_escape) {
		  putKcp(s, cdbptr-2, FALSE);
		  printf("<C>");
		  putKcp(s_cdebeg, s_cdebeg + strlen(s_cdebeg) - 1, TRUE);
		  printf("<N>");
		  s = cdbendptr; 
		  incode = OTHERCODE;
		  continue;
	       }
	    }

	    comendptr =
	       expmatch(s, (comtype == STANDARD) ? l_comend : l_acmend,
			&comptr, (char *) NULL);

	    if ((cdeptr == NULL && comptr != NULL)
		|| (cdeptr != NULL && (comptr != NULL && comptr < cdeptr))) {
	       fprintf(stderr,
	         "%s: unterminated program text within comment at line %d\n",
		   progname, lineno);
	       printf("<C>");
	       s = comptr;
	       incode = NOTCODE; incomm = TRUE;
	       continue;
	    }
	    
	    if (cdeendptr != NULL) {
	       if ((strptr  == NULL || cdeptr < strptr)
		   && (chrptr  == NULL || cdeptr < chrptr)
		   && (blksptr == NULL || cdeptr < blksptr)
		   && (blkeptr == NULL || cdeptr < blkeptr)) {
		  if (incode == INITCODE && cdeptr == s) {
		     printf("<C>");
		     putKcp(s_cdebeg, s_cdebeg + strlen(s_cdebeg) - 1, TRUE);
		  } else {
		     putKcp(s, cdeptr-1, FALSE);
		     printf("<N>");
		  }
		  s = cdeendptr;
		  incode = NOTCODE; incomm = TRUE;
		  continue;
	       }
	    } else if (strptr == NULL && chrptr == NULL
		       && blksptr == NULL && blkeptr == NULL) {
	       cdeptr = s;
	       s += strlen(s);
	       putKcp(cdeptr, s-1, FALSE);
	       incode = OTHERCODE;
	       continue;
	    } /* else there is a string/char/block on this line */

	    incode = OTHERCODE;
	 } /* if (incode) */

	 /* start of a comment? */
	 if (comptr != NULL
	     && (strptr  == NULL || comptr < strptr)
	     && (acmptr  == NULL || comptr < acmptr)
	     && (chrptr  == NULL || comptr < chrptr)
	     && (blksptr == NULL || comptr < blksptr)
	     && (blkeptr == NULL || comptr < blkeptr)) {
	    putKcp(s, comptr-1, FALSE);
	    printf("<C>");
	    s = comendptr;
	    putKcp(comptr, comendptr-1, FALSE);
	    incomm = TRUE;
	    comtype = STANDARD;
	    continue;
	 }
	 
	 /* start of an alternate-form comment? */
	 if (acmptr != NULL
	     && (strptr  == NULL || acmptr < strptr)
	     && (chrptr  == NULL || acmptr < chrptr)
	     && (blksptr == NULL || acmptr < blksptr)
	     && (blkeptr == NULL || acmptr < blkeptr)) {
	    putKcp(s, acmptr-1, FALSE);
	    printf("<C>");
	    s = acmendptr;
	    putKcp(acmptr, acmendptr, FALSE);
	    incomm = TRUE;
	    comtype = ALTERNATE;
	    continue;
	 }
	 
	 /* start of a string? */
	 if (strptr != NULL
	     && (chrptr  == NULL || strptr < chrptr)
	     && (blksptr == NULL || strptr < blksptr)
	     && (blkeptr == NULL || strptr < blkeptr)) {
	    putKcp(s, strptr-1, FALSE);
	    printf("<S>");
	    s = strendptr;
	    putKcp(strptr, strendptr-1, FALSE);
	    instr = TRUE;
	    continue;
	 }
	 
	 /* start of a character string? */
	 if (chrptr != NULL
	     && (blksptr == NULL || chrptr < blksptr)
	     && (blkeptr == NULL || chrptr < blkeptr)) {
	    putKcp(s, chrptr-1, FALSE);
	    printf("<A>");
	    s = chrendptr;
	    putKcp(chrptr, chrendptr-1, FALSE);
	    inchr = TRUE;
	    continue;
	 }
	 
	 /* end of a lexical block */
	 if (blkeptr != NULL) {
	    if (blksptr == NULL || blkeptr < blksptr) {
	       putKcp(s, blkeendptr - 1, FALSE);
	       s = blkeendptr;
	       blklevel--;
	       if (psptr >= 0 && plstack[psptr] >= blklevel) {
		  
		  /* end of current procedure */
		  blklevel = plstack[psptr];
		  
		  /* see if we should print the last proc name */
		  if (--psptr >= 0)
		     prccont = TRUE;
		  else
		     psptr = -1;
	       }
	       continue;
	    }
	 }
	 
	 /* start of a lexical block */
	 if (blksptr != NULL) {
	    putKcp(s, blksendptr - 1, FALSE);
	    s = blksendptr;
	    blklevel++;
	    continue;
	 }
	 
	 /* check for end of comment */
      } else if (incomm) {

	 cdeendptr = expmatch(s, l_cdebeg, &cdeptr, (char *) NULL);
	 texendptr = expmatch(s, l_texbeg, &texptr, (char *) NULL);
	 txmendptr = expmatch(s, l_txmbeg, &txmptr, (char *) NULL);
	 tt_endptr = expmatch(s, l_tt_beg, &tt_ptr, (char *) NULL);

	 if (code_cmnts) {

	    /* Check for program text within comment */
	    if (cdeptr != NULL
		&& (texptr == NULL || cdeptr < texptr)
		&& (tt_ptr == NULL || cdeptr < tt_ptr)
		&& (txmptr == NULL || cdeptr < txmptr)) {
	       putKcp(s, cdeptr-1, TRUE);
	       printf("<N>");
	       s = cdeendptr;
	       incode = INITCODE; incomm = FALSE;
	       continue;
	    }
	    
	    /* Check for Frame text within comment */
	    if (texptr != NULL
		&& (tt_ptr == NULL || texptr < tt_ptr)
		&& (txmptr == NULL || texptr < txmptr)) {
	       putKcp(s, texptr-1, TRUE);
	       s = texendptr;
	       if ((texendptr =
		    expmatch(s, l_texend, &texptr, (char *) NULL)) != NULL) {
		  putVcp(s, texptr-1);
		  s = texendptr;
	       } else {
		  fprintf(stderr,
		    "%s: TeX text within a comment must all be on one line\n",
		      progname);
		  s += strlen(s);
	       }
	       continue;
	    }
	    
	    /* Check for typewriter text within comment */
	    if (tt_ptr != NULL
		&& (txmptr == NULL || tt_ptr < txmptr)) {
	       putKcp(s, tt_ptr-1, TRUE);
	       s = tt_endptr;
	       if ((tt_endptr =
		    expmatch(s, l_tt_end, &tt_ptr, (char *) NULL)) != NULL) {
		  printf("<F>");
		  putKcp(s, tt_ptr-1, TRUE);
		  printf("<N>");
		  s = tt_endptr;
	       } else {
		  fprintf(stderr,
		    "%s: typewriter text within a comment must all be on one line\n\t%s",
		      progname, s);
		  s += strlen(s);
	       }
	       continue;
	    }
	    
	    /* Check for Frame math within comment */
	    if (txmptr != NULL) {
	       putKcp(s, txmptr-1, TRUE);
	       s = txmendptr;
	       if ((txmendptr =
		    expmatch(s, l_txmend, &txmptr, (char *) NULL)) != NULL) {
		  putVcp(s, txmptr-1);
		  s = txmendptr;
	       } else {
		  fprintf(stderr,
		    "%s: math within a comment must all be on one line\n",
		      progname);
		  s += strlen(s);
	       }
	       continue;
	    }
	 } /* if (code_cmnts) */

	 if ((comendptr =
	        expmatch(s, (comtype == STANDARD) ? l_comend : l_acmend,
			 (char **) NULL, (char *) NULL)) != NULL) {
	    putKcp(s, comendptr-1, TRUE);
	    s = comendptr;
	    incomm = FALSE;
	    printf("<N>");
	 } else {
	    comptr = s;
	    s += strlen(s);
	    putKcp(comptr, s-1, TRUE);
	 }
	 continue;
	 
	 /* check for end of string */
      } else if (instr) {
	 if ((strendptr =
	      expmatch(s, l_strend, (char **) NULL, (char *) NULL)) != NULL) {
	    putKcp(s, strendptr-1, TRUE);
	    s = strendptr;
	    instr = FALSE;
	    printf("<N>");
	 } else {
	    strptr = s;
	    s += strlen(s);
	    putKcp(strptr, s-1, TRUE);
	 }
	 continue;
	 
	 /* check for end of character string */
      } else if (inchr) {
	 if ((chrendptr =
	      expmatch(s, l_chrend, (char **) NULL, (char *) NULL)) != NULL) {
	    putKcp(s, chrendptr-1, TRUE);
	    s = chrendptr;
	    inchr = FALSE;
	    printf("<N>");
	 } else {
	    chrptr = s;
	    s += strlen(s);
	    putKcp(chrptr, s-1, TRUE);
	 }
	 continue;
      }
      
      /* print out the line */
      chrptr = s;
      s += strlen(s);
      putKcp(chrptr, s-1, FALSE);
      
   } while (*s);
}



/*
 * islidchr() is TRUE for things that can begin identifiers;
 * isidchr is TRUE of identifier constituents.
 */
#define islidchr(c)(isalpha(c) || (c) == '_' || (c) == '$')
#define isidchr(c) (isalnum(c) || (c) == '_' || (c) == '$')

/*
 * Write out a portion of the line
 */
   void
putKcp(start, end, nix)
   unsigned char *start;	/* Start of string to write		*/
   unsigned char *end;		/* End of string to write		*/
   boolean nix;		/* Don't look for identifiers, numbers	*/
{
   int i;
   unsigned char c;
   
   while (start <= end) {
      c = *start++;
      if (c == ' ' || c == '\t') {
         if (c == ' ') {
	    gap++; spacecol++;			/* to next column */
	 }
	 else {
	   i = ((spacecol / 8) + 1) * 8;	/* round to next tab pos */
	   gap += i - spacecol; spacecol = i;
	 }
	 continue;
      }
      else {
        if (gap) {
	   if (gap == 1 && !linehead) {		/* one space if not first */
	      outchar(' ');
	   }
	   else {				/* output tabs */
	     i = (spacecol - tabcol + 1) / 2;
	     tabcol = spacecol;
	     while (i--)
	       printf("<T>");
	   }
	   gap = 0;
	}
	/*
	  * First split off numbers.  We have a rather ad hoc
	  * definition:  A number is a digit followed by any number
	  * of digits or letters, and periods.
	  * This produces meaningless parses --- .2 is parsed as .
	  * followed by the number 2 --- but should produce decent
	  * results for most languages (outside of maybe FORTRAN and DCL).
	  */
	if (!nix) {
	    if (isdigit(c)) {
	      do {
		  if (c == 'l')
		    printf("l");
		  else
		    outchar(c);
		  spacecol++;
		  c = *start++;
	      } while (isalnum(c) || c == '.');
	      start--;
	      continue;
	    } else if (c == '#' || islidchr(c)) {
	      i = getid(--start);
	      if (i > 0)
		  printf("<K>");
	      else {
		  printf("<V>");
		  i = -i;
	      }
	      while (--i >= 0) {
		  c = *start++;
		  outchar(c);
		  spacecol++;
	      }
	      printf("<N>");
	      continue;
	    }
	}
	outchar(c);
	spacecol++;
      } /* end of else */
   } /* end of while */
}

/*
 * Write out a portion of the line verbatim
 */
   void
putVcp(start, end)
   unsigned char *start;	/* Start of string to write	*/
   unsigned char *end;		/* End of string to write	*/
{
   for ( ; start <= end; start++) {
      putchar(*start);
   }
}



/*
 * Calculate the width of a string, including tabs
 */
   int
width(s, os)
   register unsigned char *s, *os;
{
   register int i = 0;
   unsigned char c;
   
   while (s < os) {
      c = *s++;
      if (c == '\t') {
	 i = (i + 8) &~ 7;
	 continue;
      }
      if (c < ' ')
	 i += 2;
      else
	 i++;
   }
   return i;
}



/*
 * Output a string, escaping special characters
 */
   void
putstr(cp)
   register unsigned char *cp;
{
   register unsigned char c;
	
   if (cp == NULL)
      return;

   while ((c = *cp++) != '\0')
      outchar(c);
}

/*
 * The following table converts ASCII characters to a printed
 * representation, taking care of all the Frame quoting.
 *
 * Notes:
 *  Some pairs of characters are handled specially within outchar();
 *  this table contains no indication when that happens.
 *  ` is output as {`} to avoid various ligatures, such as !`.
 *  The '' and -- (and ---) ligatures are not a problem
 *  because those characters are output as macros anyway.  Using {`}
 *  is sufficient since nothing we put out will ever be given to the
 *  hyphenation algorithms anyway.
 *
 *  Some characters cannot be translated for Frame.
 */

unsigned char *printtab[256] = {
  "\\",    "^A",    "^B",    "^C",    "^D",   "^E",     "^F",    "^G",
  "^H",    "\t",    "<HardReturn>\n", "^K", "^L", "^M", "^N",    "^O",
  "^P",    "^Q",    "^R",    "^S",    "^T",    "^U",    "^V",    "^W",
  "^X",    "^Y",    "^Z",    "^[",    "^!",    "^]",    "^^",    "^_",
  " ",     "!",     "\"",    "#",     "$",     "%",     "&",     "'",
  "(",     ")",     "*",     "+",     ",",     "-",     ".",     "/",
  "0",     "1",     "2",     "3",     "4",     "5",     "6",     "7",
  "8",     "9",     ":",     ";",     "\\<",   "=",     "\\>",   "?",
  "@",     "A",     "B",     "C",     "D",     "E",     "F",     "G",
  "H",     "I",     "J",     "K",     "L",     "M",     "N",     "O",
  "P",     "Q",     "R",     "S",     "T",     "U",     "V",     "W",
  "X",     "Y",     "Z",     "[",     "\\\\",  "]",     "^",     "_",
  "`",     "a",     "b",     "c",     "d",     "e",     "f",     "g",
  "h",     "i",     "j",     "k",     "l",     "m",     "n",     "o",
  "p",     "q",     "r",     "s",     "t",     "u",     "v",     "w",
  "x",     "y",     "z",     "{",     "<M>|<N>", "}",   "~",     "^?",
  /*** this is for translation using NEXTSTEP coding ***/
  "?",     "\313",  "\347",  "\345",  "\314",  "\200",  "\201",  "\202",
  "\351",  "\203",  "\346",  "\350",  "\355",  "\352",  "\353",  "\354",
  "?",     "\204",  "\361",  "\356",  "\357",  "\315",  "\205",  "\364",
  "\362",  "\363",  "\206",  "?",     "?",     "?",     "?",     "?",   
  "\251",  "\301",  "\242",  "\243",  "\332",  "\264",  "\304",  "\244",
  "\333",  "?",     "\322",  "\307",  "\334",  "\335",  "\336",  "\337",
  "\250",  "\320",  "\240",  "\340",  "\341",  "?",     "\246",  "\245",
  "\342",  "\343",  "\323",  "\310",  "\311",  "\344",  "\302",  "\300",
  "?",     "?",     "\253",  "\366",  "\367",  "\370",  "\371",  "\372",
  "\254",  "?",     "\373",  "\374",  "?",     "\375",  "\376",  "\377",
  "\321",  "?",     "?",     "?",     "?",     "\210",  "\207",  "\211",
  "\213",  "\212",  "\214",  "\215",  "\217",  "\216",  "\220",  "\221",
  "\223",  "\256",  "\222",  "\273",  "\224",  "\225",  "?",     "\226",
  "?",     "\257",  "\316",  "\274",  "\230",  "\227",  "\231",  "\233",
  "\232",  "\276",  "\235",  "\234",  "\236",  "\365",  "\237",  "?",   
  "?",     "\277",  "\317",  "\247",  "?",     "\330",  "?",     "?",   
  /*** this is a generic identity translation
  "\200",  "\201",  "\202",  "\203",  "\204",  "\205",  "\206",  "\207",
  "\210",  "\211",  "\212",  "\213",  "\214",  "\215",  "\216",  "\217",
  "\220",  "\221",  "\222",  "\223",  "\224",  "\225",  "\226",  "\227",
  "\230",  "\231",  "\232",  "\233",  "\234",  "\235",  "\236",  "\237",
  "\240",  "\241",  "\242",  "\243",  "\244",  "\245",  "\246",  "\247",
  "\250",  "\251",  "\252",  "\253",  "\254",  "\255",  "\256",  "\257",
  "\260",  "\261",  "\262",  "\263",  "\264",  "\265",  "\266",  "\267",
  "\270",  "\271",  "\272",  "\273",  "\274",  "\275",  "\276",  "\277",
  "\300",  "\301",  "\302",  "\303",  "\304",  "\305",  "\306",  "\307",
  "\310",  "\311",  "\312",  "\313",  "\314",  "\315",  "\316",  "\317",
  "\320",  "\321",  "\322",  "\323",  "\324",  "\325",  "\326",  "\327",
  "\330",  "\331",  "\332",  "\333",  "\334",  "\335",  "\336",  "\337",
  "\340",  "\341",  "\342",  "\343",  "\344",  "\345",  "\346",  "\347",
  "\350",  "\351",  "\352",  "\353",  "\354",  "\355",  "\356",  "\357",
  "\360",  "\361",  "\362",  "\363",  "\364",  "\365",  "\366",  "\367",
  "\370",  "\371",  "\372",  "\373",  "\374",  "\375",  "\376",  "\377",
  ***/
};

#define EMDASH (-1)
/*
 * Output one character, fixed up as required.  Since there is no call-back
 * to tell outchar() to flush what it has stored, it must always produce
 * some output --- it can't simply retain state.  This makes certain
 * translations impossible, but we can live without them for now....
 */ 
   void
outchar(c)
   unsigned char c;
{
   if ((linehead || !incomm) && c == ' ') {
      putchar(' ');
      goto fin;
   }
   linehead = FALSE;
   switch (lastout) {
   case ' ':
#if 0	/* gvr hates this trick */
      if (incomm && c == '-') {
	 printf("\\xD0 ");
	 c = EMDASH;	/* For future cleverness...	*/
	 goto fin;
      }
#endif
      break;
   /* The following cases deal with special consecutive character pairs */
   case '.':
   case '|':
      if (c == lastout)
	 printf("");
      break;
   case '[':
      if (c == ']')
	 printf("");
      break;
   case '-':
      if (c == '>')
	 printf("");
      break;
   case '<':
      if (c == '-')
	 printf("");
      break;
   }
   if (incomm && c == '-') {
      putchar('-');
      goto fin;
   }
   printtab[c][1] ? printf("%s", printtab[c]) : putchar(printtab[c][0]);
 fin:
   lastout = c;
}

/*
 *	Look for a procedure beginning on this line
 */
   boolean
isproc(s)
   unsigned char *s;
{
   pname[0] = '\0';
   if ((!l_toplex || blklevel == 0)
       && expmatch(s, l_prcbeg, (char **) NULL, pname) != NULL)
      return TRUE;
   return FALSE;
}

/*
 * Get the identifier starting at s.  It is assumed that s may indeed
 * start an identifier.  Value is >0 for a keyword --- the count of
 * characters in the keyword --- and <0 if not a keyword, in which
 * case the value returned is the negative of the number of bytes in
 * the matched identifier.
 *
 * This function checks nokeyw and won't check to see if the
 * identifier found is a keyword if it is TRUE.
 */
   int
getid(s)
   unsigned char *s;
{
   unsigned char **ss = l_keywds;
   int	i = 1;
   unsigned char *cp = s;
   unsigned char firstc = *s;
   
   while (++cp, isidchr(*cp))
      i++;
   if (nokeyw)
      return -i;
   while (cp = *ss++) {
      if (!l_onecase && firstc != *cp)
	 continue;
      if ((*re_strncmp)(s, cp, i) == 0 && !isidchr(cp[i]))
	 return i;
   }
   return -i;
}



/*
 * getredirection() is intended to aid in porting C programs
 * to VMS (Vax-11 C) which does not support > and <
 * I/O redirection.  With suitable modification, it may
 * useful for other portability problems as well.
 *
 * Modified, 24-Jan-86 by Jerry Leichter
 *	When creating a new output file, force the maximum record size to
 *	512; otherwise, it ends up as 0 (though the C I/O system won't write
 *	a record longer than 512 bytes anyway) which will cause problems if
 *	the file is later opened for APPEND --- if the maximum record size
 *	is 0, C will use the length of the longest record written to the file
 *	for its buffer!
 */

#ifdef	vms
#  include	<stdio.h>
#  include	<errno.h>

   int
getredirection(argc, argv)
   int	argc;
   char	**argv;
/*
 * Process VMS redirection args.  Exit if any error is seen.
 * If getredirection() processes an argument, it is erased
 * from the vector.  getredirection() returns a new argc value.
 *
 * Warning: do not try to simplify the code for VMS.  The code
 * presupposes that getredirection() is called before any data is
 * read from stdin or written to stdout.
 *
 * Normal usage is as follows:
 *
 *	main(argc, argv)
 *	int		argc;
 *	char		*argv[];
 *	{
 *		argc = getredirection(argc, argv);
 *	}
 */
{
   register char	*ap;	/* Argument pointer */
   int			i;	/* argv[] index */
   int			j;	/* Output index */
   int			file;	/* File_descriptor */
   
   for (j = i = 1; i < argc; i++) {   /* Do all arguments */
      switch (*(ap = argv[i])) {
      case '<':			/* <file */
	 if (freopen(++ap, "r", stdin) == NULL) {
	    perror(ap);		/* Can't find file */
	    exit(errno);	/* Is a fatal error */
	 }
	 break;
	 
      case '>':			/* >file or >>file */
	 if (*++ap == '>') {	/* >>file */
	    /*
	     * If the file exists, and is writable by us,
	     * call freopen() to append to the file (using the
	     * file's current attributes).  Otherwise, create
	     * a new file with "vanilla" attributes as if
	     * the argument was given as >filename.
	     * access(name, 2) is TRUE if we can write on
	     * the specified file.
	     */
	    if (access(++ap, 2) == 0) {
	       if (freopen(ap, "a", stdout) != NULL)
		  break;	/* Exit case statement */
	       perror(ap);	/* Error, can't append */
	       exit(errno);	/* After access test */
	    }			/* If file accessable */
	 }
	 /*
	  * On VMS, we want to create the file using "standard"
	  * record attributes.  create(...) creates the file
	  * using the caller's default protection mask and
	  * "variable length, implied carriage return"
	  * attributes.  dup2() associates the file with stdout.
	  */
	 if ((file = creat(ap, 0, "rat=cr", "rfm=var", "mrs=512")) == -1
	     || dup2(file, fileno(stdout)) == -1) {
	    perror(ap);		/* Can't create file	*/
	    exit(errno);	/* is a fatal error	*/
	 }			/* If > creation	*/
	 break;			/* Exit case test	*/
	 
      default:
	 argv[j++] = ap;	/* Not a redirector	*/
	 break;			/* Exit case test	*/
      }
   }				/* For all arguments	*/
   argv[j] = NULL;		/* Terminate argv[]	*/
   return j;			/* Return new argc	*/
}

#else	/* !vms */

getredirection(argc, argv)
   int		argc;
   char		*argv[];
   /*
    * Dummy routine.
    */
{
   return (argv[0], argc);
}

#endif	/* !vms */
